function [sigmaa1,gvnew,hvnew,bvnew,z1,w1,db1,gunew,hunew,bunew,p1,q1]=structured_QZ_ds(sigmaa,gv,hv,bv,z,w,db,gu,hu,bu,p,q,kin,kfin,initialQ)
% Performs a structured double shift QZ sweep
% between indices kin and kfin
% working on quasiseparable generators.
% Last modification on June 16, 2016.
% Cleaned up w.r.t. to _old version
% (in particular: doesn't keep track of Q and Z;
% other variables have been eliminated too).

% Initialize variables.
N=length(z);
sigmav=zeros(N-1,1);
z1=zeros(N,1);
w1=zeros(N,1);
p1=zeros(N,1);
q1=zeros(N,1);
db1=zeros(N,1);
du1=zeros(N,1);


dv1=zeros(N,1);
gv1=zeros(N,4);
hv1=zeros(4,N);
bv1=zeros(4,4,N-1);
gu1=zeros(N-1,3);
hu1=zeros(3,N);
bu1=zeros(3,3,N-1);

bunew=zeros(2,2,N-1);
gunew=zeros(N-1,1);
hunew=zeros(1,N);

if kin>1
    Q=eye(3);
end
if kin==1
    Q=initialQ;
end

sigmav(1)=sigmaa(1)+z(2)*w(1);
sigmav(2)=sigmaa(2)+z(3)*w(2);

% (b)
V=zeros(3,4);
V(1,1)=gv(1,:)*hv(:,1);
V(1,2)=gv(1,:)*bv(:,:,1)*hv(:,2);
V(1,3:4)=gv(1,:)*bv(:,:,1)*bv(:,:,2);
V(2,1)=sigmav(1);
V(2,2)=gv(2,:)*hv(:,2);
V(2,3:4)=gv(2,:)*bv(:,:,2);
V(3,1)=z(3)*w(1);
V(3,2)=sigmav(2);
V(3,3:4)=gv(3,:);

auxm=Q'*V;
gv1(1,:)=auxm(1,:);
betav=auxm(2:3,:);

fv=betav(:,1:2);
phiv=betav(:,3:4);

% (c)

auxv=Q'*z(1:3);  % vector of length 3
z1(1)=auxv(1);
chi=auxv(2:3);
gamma=w(1:2);
fa=fv-chi*gamma';
phia=phiv;

% (d)
c=p(1:2);
%theta1=q(1);
theta2=q(1:2);
du(1)=db(1)+p(1)*q(1);
du(2)=db(2)+p(2)*q(2);
fu=[du(1),gu(1)*hu(2);p(2)*q(1),du(2)];
phiu=[gu(1)*bu(2);gu(2)];
epsilon=gu(1)*hu(2)-p(1)*q(2);
% fb=[du(1),epsilon;0,du(2)];
fb=[db(1),epsilon;0,db(2)]; % modified wrt manuscript
phib=phiu;

% *********************
% Chasing the Bulge
% *********************
sigmaa1=zeros(N-1,1);
for k=1:N-3  % this N may be changed later on
    
    % (a)
    epsilonb=phib*hu(k+2)-c*q(k+2); % 2x1 vector
    Phi=Q'*[fb,epsilonb;zeros(1,2),db(k+2)];
    Z=hh23(Phi(2:3,:));
    if kin>k  
        Z=eye(3);
    end
    % ****************
    % these two lines inserted later, hoping to avoid cancellation in db
    Psi=Phi*Z;
    db1(k)=Psi(1,1);
    % ****************
    
    % (c) (before (b) because of updated Q and Z)
    du(k+2)=db(k+2)+p(k+2)*q(k+2);
    Utilde=[fu,phiu*hu(k+2),phiu*bu(k+2);p(k+2)*theta2',du(k+2),gu(k+2)];
    auxm=Q'*Utilde*[Z zeros(3,1);zeros(1,3),1];
    du1(k)=auxm(1,1);
    gu1(k,:)=auxm(1,2:4);
    betau=auxm(2:3,2:4);
    fu=betau(:,1:2);
    phiu=betau(:,3);
    auxm2=[eye(2),zeros(2);zeros(1,2),hu(k+2),bu(k+2)]*[Z,zeros(3,1);zeros(1,3),1];
    hu1(:,k)=auxm2(:,1);
    bu1(:,:,k)=auxm2(:,2:4);
    
    auxv2=Q'*[c;p(k+2)];
    auxv3=Z'*[theta2;q(k+2)];
    p1(k)=auxv2(1);
    c=auxv2(2:3);
    q1(k)=auxv3(1);
    theta2=auxv3(2:3);
    
    fb=fu-c*theta2';
    phib=phiu;
    
    % (b)
    epsilona=phia*hv(:,k+2)-chi*w(k+2);
    Omega=[fa,epsilona;zeros(1,2),sigmaa(k+2)]*Z;
    % inserted these special cases so that deflated form is preserved
    if k==kfin-2
        [Q2,~]=planerot(Omega(1:2,1));
        Q=[Q2,[0;0];[0,0,1]];
        % y=[y2;0];
    elseif k>=kfin-1
        Q=eye(3);
        % y=Omega(:,1);
    else
        [Q,~]=hh3(Omega(:,1));
    end
    
    Q=Q';
    
    if kin>k+1
        Q=eye(3);
    end
    if kin==k+1
        Q=initialQ;
    end
    
    y=Q'*Omega(:,1);
    sigmaa1(k)=y(1);

    
    % (d)
    sigmav(k+2)=sigmaa(k+2)+z(k+3)*w(k+2);
    Vtilde=[fv, phiv*hv(:,k+2),phiv*bv(:,:,k+2);...
        z(k+3)*gamma', sigmav(k+2), gv(k+3,:)];
    auxm=Q'*Vtilde*[Z,zeros(3,2);zeros(2,3),eye(2)];
    dv1(k+1)=auxm(1,1);
    gv1(k+1,:)=auxm(1,2:5);
    betav=auxm(2:3,2:5);
    fv=betav(:,1:2);
    phiv=betav(:,3:4);
    auxm2=[eye(2),zeros(2,3);zeros(2),hv(:,k+2),bv(:,:,k+2)]*[Z,zeros(3,2);zeros(2,3),eye(2)];
    hv1(:,k)=auxm2(:,1);
    bv1(:,:,k)=auxm2(:,2:5);
    
    auxv=Q'*[chi;z(k+3)];
    auxv2=Z'*[gamma;w(k+2)];
    z1(k+1)=auxv(1);
    chi=auxv(2:3);
    w1(k)=auxv2(1);
    gamma=auxv2(2:3);
    
    fa=fv-chi*gamma';
    phia=phiv;
    
end

% *******************
% Last Steps
% *******************

% 1.3.1
epsilonb_nm1=phib*hu(N)-c*q(N);  % 2x1
Phi=Q'*[fb,epsilonb_nm1;zeros(1,2),db(N)];
Z=hh23(Phi(2:3,:));
if kin>N-2
    Z=eye(3);
end
% added later to prevent cancellation in diagonal of B
% ********************
Psi_Nm2=Phi*Z;
db1(N-2)=Psi_Nm2(1,1);
% ********************

%1.3.2
epsilona_N=phia*hv(:,N)-chi*w(N);
Omega_Nm2=[fa,epsilona_N]*Z;
[Q_Nm1,y_Nm1]=planerot(Omega_Nm2(:,1));
Q_Nm1=Q_Nm1';
sigmaa1(N-2)=y_Nm1(1);
VtildeN=[fv, phiv*hv(:,N)];
auxm23=Q_Nm1'*VtildeN*Z;
dv1(N-1)=auxm23(1,1);
gv1(N-1,:)=[auxm23(1,2:3),zeros(1,2)];
betav_Np1=auxm23(2,2:3);

% 1.3.3
du(N)=db(N)+p(N)*q(N)';
UtildeN=[fu,phiu*hu(N);p(N)*theta2',du(N)];
auxmN=Q'*UtildeN*Z;
du1(N-2)=auxmN(1,1);
gu1(N-2,:)=[auxmN(1,2:3),0];
fu=auxmN(2:3,2:3);
auxmN2=[eye(2),zeros(2,1);zeros(1,2),hu(N)]*Z;
hu1(:,N-2)=auxmN2(:,1);
bu1(:,:,N-2)=[auxmN2(:,2:3),zeros(3,1)];

auxv=Q_Nm1'*chi;
auxv2=Z'*[gamma;w(N)];
z1(N-1)=auxv(1);
z1(N)=auxv(2);
w1(N-2)=auxv2(1);
gamma=auxv2(2:3);
auxv2=Q'*[c;p(N)];
auxv3=Z'*[theta2;q(N)];
p1(N-2)=auxv2(1);
c=auxv2(2:3);
q1(N-2)=auxv3(1);
theta2=auxv3(2:3);

fb=fu-c*theta2';

% auxmNp1=[eye(2),zeros(2,1);zeros(1,2),hv(1,N)]*Z(:,:,N-2); % check that hv(2,N)=0
auxmNp1=[eye(2),zeros(2,1);zeros(2),hv(:,N)]*Z;
% hvtilde(:,N+1)=[auxmNp1(:,1);0];
hv1(:,N-2)=auxmNp1(:,1);
% bvtilde(:,:,N+1)=[auxmNp1(:,2:3),zeros(3,2);zeros(1,4)];
bv1(:,:,N-2)=[auxmNp1(:,2:3),zeros(4,2)];

% 1.4
Phi_Nm1=Q_Nm1'*fb;
Z_Nm1=(planerot(Phi_Nm1(2,:)'))';
Z_Nm1=fliplr(Z_Nm1);
% added later to prevent cancellation in diagonal of B
% ********************
Psi_Nm1=Phi_Nm1*Z_Nm1;
db1(N-1)=Psi_Nm1(1,1);
db1(N)=Psi_Nm1(2,2);
% ********************

hu1(:,N-1)=[Z_Nm1(:,1);0];
bu1(:,:,N-1)=[Z_Nm1(:,2),zeros(2,2);zeros(1,3)];
hu1(:,N)=[1;0;0];

hv1(:,N-1)=[Z_Nm1(:,1);zeros(2,1)];
bv1(:,:,N-1)=[Z_Nm1(:,2),zeros(2,3);zeros(2,4)];
hv1(:,N)=[1;0;0;0];

aux648=Q_Nm1'*Omega_Nm2*[[1;0;0] [0 0; Z_Nm1]];
sigmaa1(N-1)=aux648(2,2);  % MODIFIED w.r.t. manuscript
dv1(N)=betav_Np1*Z_Nm1(:,1); % ALSO MODIFIED
gv1(N,:)=[betav_Np1*Z_Nm1(:,2),0,0,0];

aux649=Q_Nm1'*fu*Z_Nm1;
du1(N-1)=aux649(1,1);
du1(N)=aux649(2,2);
gu1(N-1,:)=[aux649(1,2),0,0];

auxv=Z_Nm1'*gamma;
w1(N-1)=auxv(1);
w1(N)=auxv(2);

auxv=Q_Nm1'*c;
p1(N-1)=auxv(1);
p1(N)=auxv(2);

auxv=Z_Nm1'*theta2;
q1(N-1)=auxv(1);
q1(N)=auxv(2);



% ****************************
% Compression
% (from structured_qz_infin)
% ****************************

%  compression of V

bs=zeros(2,2,N-1);
gs=zeros(N,2);
hs=zeros(2,N);

Sold=gv1(1,:);
gs(1,1)=1;
hs(:,1)=Sold*hv1(:,1);  % this settles the (1,1) entry

gs(2,:)=[0,1];
bs(:,:,1)=[1,0; 0,0];
Snew=[Sold*bv1(:,:,1);gv1(2,:)];
hs(:,2)=Snew*hv1(:,2);

z=[gs(1,:)*hs(:,1);dv1(2)];
yold=w1(1);

for k=2:N-1
    [Vv,Y]=planerotr([yold',w1(k)']);
    ynew=Y(2);
    Aa=[z, Snew*hv1(:,k); z1(k+1)*yold', dv1(k+1)]*Vv;
    yold=ynew;
    auxvec=Aa(:,1); % to be reduced by F
    [F,~]=hh3(auxvec);
    Sold=Snew;
    Snew=[Sold*bv1(:,:,k);gv1(k+1,:)];
    Bb=F*[Aa(:,2),Snew];
    z=Bb(2:3,1);
    
    bs(:,:,k)=F(2:3,1:2)';
    gs(k+1,:)=F(2:3,3)';
    Snew=gs(k+1,:)'*gv1(k+1,:)+bs(:,:,k)'*Sold*bv1(:,:,k);
    hs(:,k+1)=Snew*hv1(:,k+1);
end
gvnew=gs;
hvnew=hs;
bvnew=bs;


%  compression of U
X=q1(1);
Y=gu1(1,:);
zed=du1(1);
%hunew(:,2)=gu1(1,:)*hu1(:,2);
hunew(2)=gu1(1,:)*hu1(:,2);
gunew(1)=1;
for k=2:N-1
    [WW,auxv]=planerotr([X',q1(k)']);
    %Zed=[zed,hunew(:,k);p1(k)*X',du1(k)]*WW;
    Zed=[zed,hunew(k);p1(k)*X',du1(k)]*WW;
    X=auxv(2);
    hp=Zed(1,2);
    hpp=Zed(2,2);
    Effe=(planerot(Zed(:,1)))'; % F
    bunew(k)=Effe(1,2);
    gunew(k)=Effe(2,2);
    Y=gunew(k)'*gu1(k,:)+bunew(k)'*Y*bu1(:,:,k);
    zed=gunew(k)'*hpp+bunew(k)'*hp;
    hunew(k+1)=Y*hu1(:,k+1);
end